'''
Linear regression and confidence intervals, for a linear regression y = k * x + d

ThH, Oct-2013
Ver 1.1
'''

import scipy.stats as ss
import numpy as np

def regress(x, y, alpha=0.05):
    '''
    Linear regression and confidence intervals, for a linear regression y = k * x + d
    
    (kd, ci) = regress(x,y,[alpha=0.05])
    
    Parameters
    ----------
    x : predictors
    y : responses
    
    Returns
    -------
    kd : coefficient estimates for a multilinear regression of the responses in y on the predictors in X.
    ci : confidence intervals
    alpha: defines the 100*(1-alpha)% confidence level in ci
    
    Note
    ----
    Similar to the MATLAB command "regress"
    
    Example
    -------
    >>> x = arange(100)
    >>> y = 0.3*x + 10 + randn(len(x))
    >>> (kd, ci) = regress.regress(x,y)
    
    '''

    stats = ss.linregress(x,y)
    kd = stats[0:2]
    se_k = stats[4]
    
    level = (1.-alpha/2.)
    tVal = ss.t.ppf(level,len(x)-2)
    ci = kd[0] + se_k*tVal*np.array([-1, 1])
    
    return(kd, ci)
    

if __name__=='__main__':
    # Test function
    data = np.genfromtxt(r'C:\Users\p20529\Coding\R\Stats\exampleData_clean.txt')
    x = data[:,0]
    y = data[:,1]
    (kd, ci) = regress(x,y)
    
    print('k = %.3f, d = %.3f, CIs = [%.3f - %.3f]' % (kd + tuple(ci)))
